@@ -27,6 +27,8 @@ ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
 
+#include <stdlib.h>
+
 #include <yara/ahocorasick.h>
 #include <yara/error.h>
 #include <yara/exec.h>
@@ -47,8 +49,8 @@ static int _yr_scanner_scan_mem_block(
     YR_MEMORY_BLOCK* block)
 {
   YR_RULES* rules = scanner->rules;
-  YR_AC_TRANSITION_TABLE transition_table = rules->ac_transition_table;
-  YR_AC_MATCH_TABLE match_table = rules->ac_match_table;
+  YR_AC_TRANSITION* transition_table = rules->ac_transition_table;
+  uint32_t* match_table = rules->ac_match_table;
 
   YR_AC_MATCH* match;
   YR_AC_TRANSITION transition;
@@ -59,30 +61,39 @@ static int _yr_scanner_scan_mem_block(
 
   while (i < block->size)
   {
-    match = match_table[state].match;
-
     if (i % 4096 == 0 && scanner->timeout > 0)
     {
-      if (yr_stopwatch_elapsed_us(&scanner->stopwatch) > scanner->timeout)
+      if (yr_stopwatch_elapsed_ns(&scanner->stopwatch) > scanner->timeout)
         return ERROR_SCAN_TIMEOUT;
     }
 
-    while (match != NULL)
+    if (match_table[state] != 0)
     {
-      if (match->backtrack <= i)
-      {
-        FAIL_ON_ERROR(yr_scan_verify_match(
-            scanner,
-            match,
-            block_data,
-            block->size,
-            block->base,
-            i - match->backtrack));
-      }
+      // If the entry corresponding to state N in the match table is zero, it
+      // means that there's no match associated to the state. If it's non-zero,
+      // its value is the 1-based index within ac_match_pool where the first
+      // match resides.
 
-      match = match->next;
+      match = &rules->ac_match_pool[match_table[state] - 1];
+
+      while (match != NULL)
+      {
+        if (match->backtrack <= i)
+        {
+          FAIL_ON_ERROR(yr_scan_verify_match(
+              scanner,
+              match,
+              block_data,
+              block->size,
+              block->base,
+              i - match->backtrack));
+        }
+
+        match = match->next;
+      }
     }
 
+
     index = block_data[i++] + 1;
     transition = transition_table[state + index];
 
@@ -103,22 +114,25 @@ static int _yr_scanner_scan_mem_block(
     state = YR_AC_NEXT_STATE(transition);
   }
 
-  match = match_table[state].match;
-
-  while (match != NULL)
+  if (match_table[state] != 0)
   {
-    if (match->backtrack <= i)
-    {
-      FAIL_ON_ERROR(yr_scan_verify_match(
-          scanner,
-          match,
-          block_data,
-          block->size,
-          block->base,
-          i - match->backtrack));
-    }
+    match = &rules->ac_match_pool[match_table[state] - 1];
 
-    match = match->next;
+    while (match != NULL)
+    {
+      if (match->backtrack <= i)
+      {
+        FAIL_ON_ERROR(yr_scan_verify_match(
+            scanner,
+            match,
+            block_data,
+            block->size,
+            block->base,
+            i - match->backtrack));
+      }
+
+      match = match->next;
+    }
   }
 
   return ERROR_SUCCESS;
@@ -128,37 +142,21 @@ static int _yr_scanner_scan_mem_block(
 static void _yr_scanner_clean_matches(
     YR_SCANNER* scanner)
 {
-  YR_RULE* rule;
-  YR_STRING** string;
+  memset(
+      scanner->rule_matches_flags, 0,
+      sizeof(YR_BITMASK) * YR_BITMASK_SIZE(scanner->rules->num_rules));
 
-  int tidx = scanner->tidx;
+  memset(
+      scanner->ns_unsatisfied_flags, 0,
+      sizeof(YR_BITMASK) * YR_BITMASK_SIZE(scanner->rules->num_namespaces));
 
-  yr_rules_foreach(scanner->rules, rule)
-  {
-    rule->t_flags[tidx] &= ~RULE_TFLAGS_MATCH;
-    rule->ns->t_flags[tidx] &= ~NAMESPACE_TFLAGS_UNSATISFIED_GLOBAL;
-  }
+  memset(
+      scanner->matches, 0,
+      sizeof(YR_MATCHES) * scanner->rules->num_strings);
 
-  if (scanner->matching_strings_arena != NULL)
-  {
-    string = (YR_STRING**) yr_arena_base_address(
-        scanner->matching_strings_arena);
-
-    while (string != NULL)
-    {
-      (*string)->matches[tidx].count = 0;
-      (*string)->matches[tidx].head = NULL;
-      (*string)->matches[tidx].tail = NULL;
-      (*string)->unconfirmed_matches[tidx].count = 0;
-      (*string)->unconfirmed_matches[tidx].head = NULL;
-      (*string)->unconfirmed_matches[tidx].tail = NULL;
-
-      string = (YR_STRING**) yr_arena_next_address(
-          scanner->matching_strings_arena,
-          string,
-          sizeof(YR_STRING*));
-    }
-  }
+  memset(
+      scanner->unconfirmed_matches, 0,
+      sizeof(YR_MATCHES) * scanner->rules->num_strings);
 }
 
 
@@ -176,7 +174,40 @@ YR_API int yr_scanner_create(
 
   FAIL_ON_ERROR_WITH_CLEANUP(
       yr_hash_table_create(64, &new_scanner->objects_table),
-      yr_scanner_destroy(new_scanner));
+      yr_free(new_scanner));
+
+  new_scanner->rules = rules;
+  new_scanner->entry_point = YR_UNDEFINED;
+  new_scanner->canary = rand();
+
+  // By default report both matching and non-matching rules.
+  new_scanner->flags = \
+      SCAN_FLAGS_REPORT_RULES_MATCHING |
+      SCAN_FLAGS_REPORT_RULES_NOT_MATCHING;
+
+  new_scanner->rule_matches_flags = (YR_BITMASK*) yr_calloc(
+      sizeof(YR_BITMASK), YR_BITMASK_SIZE(rules->num_rules));
+
+  new_scanner->ns_unsatisfied_flags = (YR_BITMASK*) yr_calloc(
+      sizeof(YR_BITMASK), YR_BITMASK_SIZE(rules->num_namespaces));
+
+  new_scanner->matches = (YR_MATCHES*) yr_calloc(
+      rules->num_strings, sizeof(YR_MATCHES));
+
+  new_scanner->unconfirmed_matches = (YR_MATCHES*) yr_calloc(
+      rules->num_strings, sizeof(YR_MATCHES));
+
+  #ifdef YR_PROFILING_ENABLED
+  new_scanner->profiling_info = yr_calloc(rules->num_rules,  sizeof(YR_PROFILING_INFO));
+
+  if (new_scanner->profiling_info == NULL)
+  {
+    yr_scanner_destroy(new_scanner);
+    return ERROR_INSUFFICIENT_MEMORY;
+  }
+  #else
+  new_scanner->profiling_info = NULL;
+  #endif
 
   external = rules->externals_list_head;
 
@@ -186,6 +217,7 @@ YR_API int yr_scanner_create(
 
     FAIL_ON_ERROR_WITH_CLEANUP(
         yr_object_from_external_variable(external, &object),
+        // cleanup
         yr_scanner_destroy(new_scanner));
 
     FAIL_ON_ERROR_WITH_CLEANUP(
@@ -194,15 +226,14 @@ YR_API int yr_scanner_create(
             external->identifier,
             NULL,
             (void*) object),
+        // cleanup
+        yr_object_destroy(object);
         yr_scanner_destroy(new_scanner));
 
+    yr_object_set_canary(object, new_scanner->canary);
     external++;
   }
 
-  new_scanner->rules = rules;
-  new_scanner->tidx = -1;
-  new_scanner->entry_point = UNDEFINED;
-
   *scanner = new_scanner;
 
   return ERROR_SUCCESS;
@@ -231,6 +262,14 @@ YR_API void yr_scanner_destroy(
         (YR_HASH_TABLE_FREE_VALUE_FUNC) yr_object_destroy);
   }
 
+  #ifdef YR_PROFILING_ENABLED
+  yr_free(scanner->profiling_info);
+  #endif
+
+  yr_free(scanner->rule_matches_flags);
+  yr_free(scanner->ns_unsatisfied_flags);
+  yr_free(scanner->matches);
+  yr_free(scanner->unconfirmed_matches);
   yr_free(scanner);
 }
 
@@ -249,14 +288,24 @@ YR_API void yr_scanner_set_timeout(
     YR_SCANNER* scanner,
     int timeout)
 {
-  scanner->timeout = timeout * 1000000L;  // convert timeout to microseconds.
+  // Convert timeout from seconds to nanoseconds.
+  scanner->timeout = timeout * 1000000000ULL;
 }
 
-
 YR_API void yr_scanner_set_flags(
     YR_SCANNER* scanner,
     int flags)
 {
+  // For backward compatibility, if neither SCAN_FLAGS_REPORT_RULES_MATCHING
+  // nor SCAN_FLAGS_REPORT_RULES_NOT_MATCHING are specified, both are assumed.
+
+  if (!(flags & SCAN_FLAGS_REPORT_RULES_MATCHING) &&
+      !(flags & SCAN_FLAGS_REPORT_RULES_NOT_MATCHING))
+  {
+    flags |= SCAN_FLAGS_REPORT_RULES_MATCHING |
+             SCAN_FLAGS_REPORT_RULES_NOT_MATCHING;
+  }
+
   scanner->flags = flags;
 }
 
@@ -338,14 +387,17 @@ YR_API int yr_scanner_scan_mem_blocks(
   YR_RULE* rule;
   YR_MEMORY_BLOCK* block;
 
-  int tidx = 0;
-  int result = ERROR_SUCCESS;
+  uint32_t max_match_data;
 
-  uint64_t elapsed_time;
+  int i, result = ERROR_SUCCESS;
 
   if (scanner->callback == NULL)
     return ERROR_CALLBACK_REQUIRED;
 
+  FAIL_ON_ERROR(yr_get_configuration(
+      YR_CONFIG_MAX_MATCH_DATA,
+      &max_match_data))
+
   scanner->iterator = iterator;
   rules = scanner->rules;
   block = iterator->first(iterator);
@@ -353,34 +405,15 @@ YR_API int yr_scanner_scan_mem_blocks(
   if (block == NULL)
     return ERROR_SUCCESS;
 
-  yr_mutex_lock(&rules->mutex);
-
-  while (tidx < YR_MAX_THREADS && YR_BITARRAY_TEST(rules->tidx_mask, tidx))
-  {
-    tidx++;
-  }
-
-  if (tidx < YR_MAX_THREADS)
-    YR_BITARRAY_SET(rules->tidx_mask, tidx);
-  else
-    result = ERROR_TOO_MANY_SCAN_THREADS;
-
-  yr_mutex_unlock(&rules->mutex);
-
-  if (result != ERROR_SUCCESS)
-    return result;
-
-  scanner->tidx = tidx;
   scanner->file_size = block->size;
 
-  yr_set_tidx(tidx);
-
-  result = yr_arena_create(1048576, 0, &scanner->matches_arena);
-
-  if (result != ERROR_SUCCESS)
-    goto _exit;
-
-  result = yr_arena_create(4096, 0, &scanner->matching_strings_arena);
+  // Create the notebook that will hold the YR_MATCH structures representing
+  // each match found. This notebook will also contain snippets of the matching
+  // data (the "data" field in YR_MATCH points to the snippet corresponding to
+  // the match). Each notebook's page can store up to 1024 matches.
+  result = yr_notebook_create(
+      1024 * (sizeof(YR_MATCH) + max_match_data),
+      &scanner->matches_notebook);
 
   if (result != ERROR_SUCCESS)
     goto _exit;
@@ -398,7 +431,7 @@ YR_API int yr_scanner_scan_mem_blocks(
       continue;
     }
 
-    if (scanner->entry_point == UNDEFINED)
+    if (scanner->entry_point == YR_UNDEFINED)
     {
       YR_TRYCATCH(
         !(scanner->flags & SCAN_FLAGS_NO_TRYCATCH),
@@ -443,23 +476,27 @@ YR_API int yr_scanner_scan_mem_blocks(
   if (result != ERROR_SUCCESS)
     goto _exit;
 
-  yr_rules_foreach(rules, rule)
+  for (i = 0, rule = rules->rules_list_head;
+       !RULE_IS_NULL(rule);
+       i++, rule++)
   {
-    int message;
+    int message = 0;
 
-    if (rule->t_flags[tidx] & RULE_TFLAGS_MATCH &&
-        !(rule->ns->t_flags[tidx] & NAMESPACE_TFLAGS_UNSATISFIED_GLOBAL))
+    if (yr_bitmask_is_set(scanner->rule_matches_flags, i) &&
+        yr_bitmask_is_not_set(scanner->ns_unsatisfied_flags, rule->ns->idx))
     {
-      message = CALLBACK_MSG_RULE_MATCHING;
+      if (scanner->flags & SCAN_FLAGS_REPORT_RULES_MATCHING)
+        message = CALLBACK_MSG_RULE_MATCHING;
     }
     else
     {
-      message = CALLBACK_MSG_RULE_NOT_MATCHING;
+      if (scanner->flags & SCAN_FLAGS_REPORT_RULES_NOT_MATCHING)
+        message = CALLBACK_MSG_RULE_NOT_MATCHING;
     }
 
-    if (!RULE_IS_PRIVATE(rule))
+    if (message != 0 && !RULE_IS_PRIVATE(rule))
     {
-      switch (scanner->callback(message, rule, scanner->user_data))
+      switch (scanner->callback(scanner, message, rule, scanner->user_data))
       {
         case CALLBACK_ABORT:
           result = ERROR_SUCCESS;
@@ -472,46 +509,22 @@ YR_API int yr_scanner_scan_mem_blocks(
     }
   }
 
-  scanner->callback(CALLBACK_MSG_SCAN_FINISHED, NULL, scanner->user_data);
+  scanner->callback(
+      scanner,
+      CALLBACK_MSG_SCAN_FINISHED,
+      NULL,
+      scanner->user_data);
 
 _exit:
 
-  elapsed_time = yr_stopwatch_elapsed_us(&scanner->stopwatch);
-
-  #ifdef PROFILING_ENABLED
-  yr_rules_foreach(rules, rule)
-  {
-    #ifdef _WIN32
-    InterlockedAdd64(&rule->time_cost, rule->time_cost_per_thread[tidx]);
-    #else
-    __sync_fetch_and_add(&rule->time_cost, rule->time_cost_per_thread[tidx]);
-    #endif
-
-    rule->time_cost_per_thread[tidx] = 0;
-  }
-  #endif
-
   _yr_scanner_clean_matches(scanner);
 
-  if (scanner->matches_arena != NULL)
+  if (scanner->matches_notebook != NULL)
   {
-    yr_arena_destroy(scanner->matches_arena);
-    scanner->matches_arena = NULL;
+    yr_notebook_destroy(scanner->matches_notebook);
+    scanner->matches_notebook = NULL;
   }
 
-  if (scanner->matching_strings_arena != NULL)
-  {
-    yr_arena_destroy(scanner->matching_strings_arena);
-    scanner->matching_strings_arena = NULL;
-  }
-
-  yr_mutex_lock(&rules->mutex);
-  YR_BITARRAY_UNSET(rules->tidx_mask, tidx);
-  rules->time_cost += elapsed_time;
-  yr_mutex_unlock(&rules->mutex);
-
-  yr_set_tidx(-1);
-
   return result;
 }
 
@@ -628,5 +641,106 @@ YR_API YR_RULE* yr_scanner_last_error_rule(
   if (scanner->last_error_string == NULL)
     return NULL;
 
-  return scanner->last_error_string->rule;
+  return &scanner->rules->rules_list_head[scanner->last_error_string->rule_idx];
+}
+
+
+static int sort_by_cost_desc(
+    const struct YR_RULE_PROFILING_INFO* r1,
+    const struct YR_RULE_PROFILING_INFO* r2)
+{
+  if (r1->cost < r2->cost)
+    return 1;
+
+  if (r1->cost > r2->cost)
+    return -1;
+
+  return 0;
+}
+
+//
+// yr_scanner_get_profiling_info
+//
+// Returns a pointer to an array of YR_RULE_PROFILING_INFO structures with
+// information about the cost of each rule. The rules are sorted by cost
+// in descending order and the last item in the array has rule == NULL.
+// The caller is responsible for freeing the returned array by calling
+// yr_free. Calling this function only makes sense if YR_PROFILING_ENABLED
+// is defined, if not, the cost for each rule won't be computed, it will be
+// set to 0 for all rules.
+//
+YR_API YR_RULE_PROFILING_INFO* yr_scanner_get_profiling_info(
+    YR_SCANNER* scanner)
+{
+  YR_RULE_PROFILING_INFO* profiling_info = yr_malloc(
+      (scanner->rules->num_rules + 1) * sizeof(YR_RULE_PROFILING_INFO));
+
+  if (profiling_info == NULL)
+    return NULL;
+
+  for (uint32_t i = 0; i < scanner->rules->num_rules; i++)
+  {
+    profiling_info[i].rule = &scanner->rules->rules_list_head[i];
+    #ifdef YR_PROFILING_ENABLED
+    profiling_info[i].cost = \
+        scanner->profiling_info[i].exec_time +
+        (scanner->profiling_info[i].atom_matches *
+         scanner->profiling_info[i].match_time) /
+        YR_MATCH_VERIFICATION_PROFILING_RATE;
+    #else
+    memset(&profiling_info[i], 0, sizeof(YR_RULE_PROFILING_INFO));
+    #endif
+  }
+
+  qsort(
+      profiling_info,
+      scanner->rules->num_rules,
+      sizeof(YR_RULE_PROFILING_INFO),
+      (int (*)(const void *, const void *)) sort_by_cost_desc);
+
+  profiling_info[scanner->rules->num_rules].rule = NULL;
+  profiling_info[scanner->rules->num_rules].cost = 0;
+
+  return profiling_info;
+}
+
+
+YR_API void yr_scanner_reset_profiling_info(
+    YR_SCANNER* scanner)
+{
+  #ifdef YR_PROFILING_ENABLED
+  memset(
+    scanner->profiling_info, 0,
+    scanner->rules->num_rules * sizeof(YR_PROFILING_INFO));
+  #endif
+}
+
+YR_API int yr_scanner_print_profiling_info(
+    YR_SCANNER* scanner)
+{
+  printf("\n===== PROFILING INFORMATION =====\n\n");
+
+  YR_RULE_PROFILING_INFO* info = yr_scanner_get_profiling_info(scanner);
+
+  if (info == NULL)
+    return ERROR_INSUFFICIENT_MEMORY;
+
+  YR_RULE_PROFILING_INFO* rpi = info;
+
+  while (rpi->rule != NULL)
+  {
+    printf(
+        "%10" PRIu64 " %s:%s: \n",
+        rpi->cost,
+        rpi->rule->ns->name,
+        rpi->rule->identifier);
+
+    rpi++;
+  }
+
+  printf("\n=================================\n");
+
+  yr_free(info);
+
+  return ERROR_SUCCESS;
 }
